// Forlix SrcDS MRRPatch
// http://forlix.org/, df@forlix.org
//
// Copyright (c) 2010-2014 Dominik Friedrichs

#include "SrcDS_MRRPatch.h"

#define PLUGIN_VERSION_BARE     "1.03"
#define PLUGIN_VERSION_CVAR     "forlix_mrrpatch_version"

#ifdef _DEBUG
  #define PLUGIN_VERSION        PLUGIN_VERSION_BARE " debug"
#else
  #define PLUGIN_VERSION        PLUGIN_VERSION_BARE
#endif

#define MRR_TRIGGERDELAY        "20"
#define MRR_MAXPLAYERS          "8"
#define MRR_MARKERFILE          "update_request.txt"

Forlix_SrcDS_MRRPatch g_Forlix_SrcDS_MRRPatch;
PLUGIN_EXPOSE(Forlix_SrcDS_MRRPatch, g_Forlix_SrcDS_MRRPatch);

int g_mrr_triggerdelay = 0;
int g_mrr_maxplayers = 0;

void triggerdelay_callback(IConVar *var, const char *pOldValue, float flOldValue);
void maxplayers_callback(IConVar *var, const char *pOldValue, float flOldValue);

void manualtrigger_callback(const CCommand &command);

ConVar g_version_cvar(PLUGIN_VERSION_CVAR, PLUGIN_VERSION,
                      FCVAR_SPONLY|FCVAR_NOTIFY|FCVAR_PRINTABLEONLY,
                      "Forlix MRRPatch plugin version");

ConVar g_triggerdelay_cvar("forlix_mrrpatch_triggerdelay", MRR_TRIGGERDELAY,
                           FCVAR_NONE,
                           "Sets the time in seconds between receiving an MRR and trying a server shutdown",
                           true, 0.0, true, 3600.0, triggerdelay_callback);

ConVar g_maxplayers_cvar("forlix_mrrpatch_maxplayers", MRR_MAXPLAYERS,
                         FCVAR_NONE,
                         "Server shutdown will be postponed until player count has dropped to this threshold",
                         true, 0.0, true, 64.0, maxplayers_callback);

ConCommand g_manualtrigger("forlix_mrrpatch_manualtrigger", manualtrigger_callback,
                          "Initiates the shutdown sequence as if an MRR has been received");

CMemPatch g_PatchMasterRequestRestart;
void Hook_MasterRequestRestart(const char *str, char c);
void MasterRequestRestartTick(bool trigger=false);

SH_DECL_HOOK3_void(IServerGameDLL, ServerActivate, SH_NOATTRIB, 0, edict_t *, int, int);
SH_DECL_HOOK1_void(IServerGameDLL, GameFrame, SH_NOATTRIB, 0, bool);
SH_DECL_HOOK6(IServerGameDLL, LevelInit, SH_NOATTRIB, 0, bool, char const *, char const *, char const *, char const *, bool, bool);
SH_DECL_HOOK0_void(IServerGameDLL, LevelShutdown, SH_NOATTRIB, 0);

void SayTextMessage(const char *msg);
void TopMessage(const char *msg,
                Color color,
                int level,
                int time);

int GetHumanPlayerCount();
int GetUserMessageIndex(const char *type);

int g_SayText;

bool g_HasMapTickedYet = false;
double g_UniversalTime = 0.0;

int g_MaxClients = 0;

IServerGameDLL *server = NULL;
IPlayerInfoManager *playerinfomanager = NULL;

IVEngineServer *engine = NULL;
ICvar *icvar = NULL;
IFileSystem *fs = NULL;
IServerPluginHelpers *serverpluginhelpers = NULL;
IServerPluginCallbacks *vsp_callbacks = NULL;
CGlobalVars *gpGlobals = NULL;


bool Forlix_SrcDS_MRRPatch::RegisterConCommandBase(ConCommandBase *pCommandBase)
{
  return(META_REGCVAR(pCommandBase));
}


void triggerdelay_callback(IConVar *var,
                           const char *pOldValue,
                           float flOldValue)
{
  g_mrr_triggerdelay = g_triggerdelay_cvar.GetInt();
  return;
}


void maxplayers_callback(IConVar *var,
                         const char *pOldValue,
                         float flOldValue)
{
  g_mrr_maxplayers = g_maxplayers_cvar.GetInt();
  return;
}


void manualtrigger_callback(const CCommand &command)
{
  META_LOG(g_PLAPI, "Received manual MasterRequestRestart trigger");

  MasterRequestRestartTick(true);
  return;
}


bool Forlix_SrcDS_MRRPatch::Load(PluginId id, ISmmAPI *ismm, char *error, size_t maxlen, bool late)
{
  PLUGIN_SAVEVARS();

  GET_V_IFACE_ANY(GetServerFactory, server, IServerGameDLL, INTERFACEVERSION_SERVERGAMEDLL);
  GET_V_IFACE_ANY(GetServerFactory, playerinfomanager, IPlayerInfoManager, INTERFACEVERSION_PLAYERINFOMANAGER);

  GET_V_IFACE_CURRENT(GetEngineFactory, engine, IVEngineServer, INTERFACEVERSION_VENGINESERVER);
  GET_V_IFACE_CURRENT(GetEngineFactory, icvar, ICvar, CVAR_INTERFACE_VERSION);
  GET_V_IFACE_CURRENT(GetFileSystemFactory, fs, IFileSystem, FILESYSTEM_INTERFACE_VERSION);
  GET_V_IFACE_CURRENT(GetEngineFactory, serverpluginhelpers, IServerPluginHelpers, INTERFACEVERSION_ISERVERPLUGINHELPERS);

  gpGlobals = g_SMAPI->GetCGlobals();

  g_pCVar = (ICvar *)((g_SMAPI->GetEngineFactory())(CVAR_INTERFACE_VERSION, NULL));
  ConVar_Register(0, this);

  if((vsp_callbacks = ismm->GetVSPInfo(NULL)) == NULL)
  {
    ismm->AddListener(this, this);
    ismm->EnableVSPListener();
  }

  SH_ADD_HOOK_MEMFUNC(IServerGameDLL, ServerActivate, server, this, &Forlix_SrcDS_MRRPatch::Hook_ServerActivate, true);
  SH_ADD_HOOK_MEMFUNC(IServerGameDLL, GameFrame, server, this, &Forlix_SrcDS_MRRPatch::Hook_GameFrame, true);
  SH_ADD_HOOK_MEMFUNC(IServerGameDLL, LevelInit, server, this, &Forlix_SrcDS_MRRPatch::Hook_LevelInit, true);
  SH_ADD_HOOK_MEMFUNC(IServerGameDLL, LevelShutdown, server, this, &Forlix_SrcDS_MRRPatch::Hook_LevelShutdown, false);

  // MasterRequestRestart (in-function address, executed when server is out of date)
  // %cMasterRequestRestart
  CSigScan masterrequestrestart("FF ?? 84 C0 74 ?? 56 8B 35 ?? ?? ?? ?? 6A 03 68 "
                                "?? ?? ?? ?? FF D6 83 C4 08 80 3D ?? ?? ?? ?? 00", "engine.dll");

  if(!masterrequestrestart.FindSignature())
    META_LOG(g_PLAPI, "Failed to locate MasterRequestRestart signature");
  else
  // patch inserts call to our hook function, replacing the original tier0.Msg call
    g_PatchMasterRequestRestart.PatchMemory("E8 ## ## ## ## 83 C4 08 B0 00 84 C0",
                                            masterrequestrestart.GetSigAddress(20), &Hook_MasterRequestRestart);

  if((g_SayText = GetUserMessageIndex("SayText")) < 0)
    META_LOG(g_PLAPI, "Failed to locate SayText message index");

  if(late)
    Hook_ServerActivate(NULL, 0, 0);

  return(true);
}


bool Forlix_SrcDS_MRRPatch::Unload(char *error, size_t maxlen)
{
  g_PatchMasterRequestRestart.RestoreMemory();

  SH_REMOVE_HOOK_MEMFUNC(IServerGameDLL, ServerActivate, server, this, &Forlix_SrcDS_MRRPatch::Hook_ServerActivate, true);
  SH_REMOVE_HOOK_MEMFUNC(IServerGameDLL, GameFrame, server, this, &Forlix_SrcDS_MRRPatch::Hook_GameFrame, true);
  SH_REMOVE_HOOK_MEMFUNC(IServerGameDLL, LevelInit, server, this, &Forlix_SrcDS_MRRPatch::Hook_LevelInit, true);
  SH_REMOVE_HOOK_MEMFUNC(IServerGameDLL, LevelShutdown, server, this, &Forlix_SrcDS_MRRPatch::Hook_LevelShutdown, false);

  return(true);
}


void Forlix_SrcDS_MRRPatch::OnVSPListening(IServerPluginCallbacks *iface)
{
  vsp_callbacks = iface;
  return;
}


void Forlix_SrcDS_MRRPatch::Hook_ServerActivate(edict_t *pEdictList,
                                                int edictCount,
                                                int clientMax)
{
  // dont use clientmax from arguments!
  g_MaxClients = gpGlobals->maxClients;

  triggerdelay_callback(NULL, NULL, 0.0);
  maxplayers_callback(NULL, NULL, 0.0);
  return;
}


void Forlix_SrcDS_MRRPatch::Hook_GameFrame(bool simulating)
{
  static float lasttickedtime = 0.0;

  if(simulating && g_HasMapTickedYet)
    g_UniversalTime += gpGlobals->curtime - lasttickedtime;
  else
    g_UniversalTime += gpGlobals->interval_per_tick;

  lasttickedtime = gpGlobals->curtime;
  g_HasMapTickedYet = true;

  MasterRequestRestartTick();
  return;
}


bool Forlix_SrcDS_MRRPatch::Hook_LevelInit(const char *pMapName,
                                           char const *pMapEntities,
                                           char const *pOldLevel,
                                           char const *pLandmarkName,
                                           bool loadGame,
                                           bool background)
{
  return(true);
}


void Forlix_SrcDS_MRRPatch::Hook_LevelShutdown()
{
  g_HasMapTickedYet = false;
  return;
}


void Hook_MasterRequestRestart(const char *str, char c)
{
  MasterRequestRestartTick(true);
  return;
}


void MasterRequestRestartTick(bool trigger)
{
  enum State {Idle=0, Lock, TryRestart, NotifyPlayers, ForceRestart};

  static State state = Idle;
  static double delay = 0.0;

  if(g_UniversalTime < delay)
  // delayed state processing
    return;

  switch(state)
  {
    case Idle:
    if(trigger)
    {
      META_LOG(g_PLAPI, "MasterRequestRestart triggered, trying restart in %u seconds",
                        g_mrr_triggerdelay);
      state = TryRestart;
      delay = g_UniversalTime + g_mrr_triggerdelay;
    }
    break;

    case Lock:
    // do nothing
    break;

    case TryRestart:
    {
      int numplayers = GetHumanPlayerCount();

      if(numplayers > g_mrr_maxplayers)
      // postpone until player count drops to threshold
      {
        META_LOG(g_PLAPI, "Restart pending, player count %u needs to drop to %u",
                          numplayers, g_mrr_maxplayers);
        delay = g_UniversalTime + 120.0;
      }
      else
      if(numplayers > 0)
      // below threshold, notify players
        state = NotifyPlayers;
      else
      // no players connected, skip notifying
        state = ForceRestart;
    }
    break;

    case NotifyPlayers:
    {
      // notification times (relative to restart, in reversed order)
      static const float notify_time[] = {10.0, 30.0};
      static int n = -1;

      if(n < 0)
      // first notification
      {
        META_LOG(g_PLAPI, "Notifying players about imminent server restart");
        n = (sizeof(notify_time)/sizeof(notify_time[0]))-1;
      }

      TopMessage("Required game update available", Color(255,64,0,255), 0, 10);

      char msg[128];
      sprintf_s(msg, sizeof(msg), " \n"
                "Restart your game to receive the latest update.\n"
                "This server will update and restart in %.f seconds!\n ", notify_time[n]);

      SayTextMessage(msg);
      delay = g_UniversalTime + notify_time[n];

      if(n-- > 0)
      // subtract relative time of next notice
        delay -= notify_time[n];
      else
      // last notice sent, restart after this delay
        state = ForceRestart;
    }
    break;

    case ForceRestart:
    {
      META_LOG(g_PLAPI, "Triggering server restart to receive the latest update");

      // place marker file
      FileHandle_t fh = fs->Open(MRR_MARKERFILE, "wt", "GAME");
      if(fh)
      {
        fs->FPrintf(fh, "MasterRequestRestart\n%s\n", TimeStampStr());
        fs->Close(fh);
      }

      state = Lock;
      engine->ServerCommand("_restart\n");
    }
    break;
  }

  return;
}


void SayTextMessage(const char *msg)
{
  if(g_SayText < 0)
    return;

  MRecipientFilter filter;
  filter.AddAllPlayers(g_MaxClients);

  bf_write *pmsg = engine->UserMessageBegin(&filter, g_SayText);

  if(pmsg)
  {
    pmsg->WriteByte(0);
    pmsg->WriteString(msg);
    engine->MessageEnd();
  }

  return;
}


void TopMessage(const char *msg,
                Color color,
                int level,
                int time)
{
  KeyValues *kv = new KeyValues("topmsg");
  kv->SetString("title", msg);
  kv->SetColor("color", color);
  kv->SetInt("level", level);
  kv->SetInt("time", time);

  for(int i = 1; i <= g_MaxClients; i++)
  {
    edict_t *pEdict = engine->PEntityOfEntIndex(i);
    if(!pEdict
    || pEdict->IsFree())
      continue;

    IPlayerInfo *playerinfo = playerinfomanager->GetPlayerInfo(pEdict);
    if(!playerinfo
    || !playerinfo->IsConnected()
    || playerinfo->IsHLTV() || playerinfo->IsFakeClient())
      continue;

    serverpluginhelpers->CreateMessage(pEdict, DIALOG_MSG, kv, vsp_callbacks);
  }

  kv->deleteThis();
  return;
}


int GetHumanPlayerCount()
{
  int players = 0;

  for(int i = 1; i <= g_MaxClients; i++)
  {
    edict_t *pEdict = engine->PEntityOfEntIndex(i);
    if(!pEdict
    || pEdict->IsFree())
      continue;

    IPlayerInfo *playerinfo = playerinfomanager->GetPlayerInfo(pEdict);
    if(!playerinfo
    || !playerinfo->IsConnected()
    || playerinfo->IsHLTV() || playerinfo->IsFakeClient())
      continue;

    players++;
  }

  return(players);
}


int GetUserMessageIndex(const char *type)
{
	int i = 0;
  int size;
  char buf[64];

  while(server->GetUserMessageInfo(i, buf, sizeof(buf), size))
  {
    if(!strcmp(buf, type))
      return(i);

    i++;
  }

  return(-1);
}


const char *Forlix_SrcDS_MRRPatch::GetLicense()
{
  return("GPL");
}

const char *Forlix_SrcDS_MRRPatch::GetVersion()
{
  return(PLUGIN_VERSION);
}

const char *Forlix_SrcDS_MRRPatch::GetDate()
{
  return(__DATE__);
}

const char *Forlix_SrcDS_MRRPatch::GetLogTag()
{
  return("Forlix_MRRPatch");
}

const char *Forlix_SrcDS_MRRPatch::GetAuthor()
{
  return("Forlix (Dominik Friedrichs)");
}

const char *Forlix_SrcDS_MRRPatch::GetDescription()
{
  return("Automatic server shutdown upon MasterRequestRestart");
}

const char *Forlix_SrcDS_MRRPatch::GetName()
{
  return("Forlix MRRPatch");
}

const char *Forlix_SrcDS_MRRPatch::GetURL()
{
  return("http://forlix.org/");
}
